"""Classes representing database entities."""

from __future__ import annotations

from collections.abc import Collection
from typing import TYPE_CHECKING, Any

from sqlfactory.statement import Raw, Statement

if TYPE_CHECKING:
    from sqlfactory.condition.in_condition import In  # pragma: nocover
    from sqlfactory.condition.simple import Eq, Ge, Gt, Le, Lt, Ne  # pragma: nocover


class Expression(Statement):
    """
    Expression as statement

    Represents expression that can be used in SQL statements. It provides basic methods for creating complex
    expressions by combining columns, functions and literals together to one arithmetic function.
    """

    def __gt__(self, other: Statement | Any) -> Gt:
        """Shorthand to produce conditional SQL statement <column> > <other>."""
        from sqlfactory.condition.simple import Gt  # pylint: disable=import-outside-toplevel

        return Gt(self, other)

    def __ge__(self, other: Statement | Any) -> Ge:
        """Shorthand to produce conditional SQL statement <column> >= <other>."""
        from sqlfactory.condition.simple import Ge  # pylint: disable=import-outside-toplevel

        return Ge(self, other)

    def __lt__(self, other: Statement | Any) -> Lt:
        """Shorthand to produce conditional SQL statement <column> < <other>."""
        from sqlfactory.condition.simple import Lt  # pylint: disable=import-outside-toplevel

        return Lt(self, other)

    def __le__(self, other: Statement | Any) -> Le:
        """Shorthand to produce conditional SQL statement <column> <= <other>."""
        from sqlfactory.condition.simple import Le  # pylint: disable=import-outside-toplevel

        return Le(self, other)

    def __eq__(self, other: Statement | Any) -> Eq:  # type: ignore[override]
        """Shorthand to produce conditional SQL statement <column> = <other>."""
        from sqlfactory.condition.simple import Eq  # pylint: disable=import-outside-toplevel

        return Eq(self, other)

    def __ne__(self, other: Statement | Any) -> Ne:  # type: ignore[override]
        """Shorthand to produce conditional SQL statement <column> != <other>."""
        from sqlfactory.condition.simple import Ne  # pylint: disable=import-outside-toplevel

        return Ne(self, other)

    def __add__(self, other: Statement | Any) -> Expression:
        return BinaryExpression(self, "+", other)

    def __sub__(self, other: Statement | Any) -> Expression:
        return BinaryExpression(self, "-", other)

    def __mul__(self, other: Statement | Any) -> Expression:
        return BinaryExpression(self, "*", other)

    def __truediv__(self, other: Statement | Any) -> Expression:
        return BinaryExpression(self, "/", other)

    def __mod__(self, other: Statement | Any) -> Expression:
        return BinaryExpression(self, "%", other)

    def __and__(self, other: Statement | Any) -> Expression:
        return BinaryExpression(self, "&", other)

    def __or__(self, other: Statement | Any) -> Expression:
        return BinaryExpression(self, "|", other)

    def __xor__(self, other: Statement | Any) -> Expression:
        return BinaryExpression(self, "^", other)

    def __lshift__(self, other: Statement | Any) -> Expression:
        return BinaryExpression(self, "<<", other)

    def __rshift__(self, other: Statement | Any) -> Expression:
        return BinaryExpression(self, ">>", other)

    def __neg__(self) -> Expression:
        return UnaryExpression("~", self)

    def in_(self, args: Collection[Any]) -> In:
        """Shorthand to `In(column, args)`."""
        from sqlfactory.condition.in_condition import In  # pylint: disable=import-outside-toplevel

        return In(self, args)

    def IN(self, args: Collection[Any]) -> In:
        # pylint: disable=invalid-name
        """Shorthand to `In(column, args)`"""
        return self.in_(args)

    def not_in(self, args: Collection[Any]) -> In:
        """Shorthand to `In(column, args, negative=True)`."""
        from sqlfactory.condition.in_condition import In  # pylint: disable=import-outside-toplevel

        return In(self, args, negative=True)

    def NOT_IN(self, args: Collection[Any]) -> In:
        # pylint: disable=invalid-name
        """Shorthand to `In(column, ..., negative=True)`."""
        return self.not_in(args)


class UnaryExpression(Expression):
    """
    Expression where operator preceeds expression.
    """

    def __init__(self, operator: str, statement: Statement | Any) -> None:
        super().__init__()

        self.operator = operator
        self.statement = statement

    def __str__(self) -> str:
        stmt = self.statement if isinstance(self.statement, Statement) else self.dialect.placeholder
        return f"({self.operator}{stmt!s})"

    @property
    def args(self) -> list[Any]:
        return self.statement.args if isinstance(self.statement, Statement) else [self.statement]


class BinaryExpression(Expression):
    """
    Expression where left and right statements are combined using operator.
    """

    def __init__(self, left: Statement | Any, operator: str, right: Statement | Any) -> None:
        super().__init__()

        self.left = left
        self.operator = operator
        self.right = right

    def __str__(self) -> str:
        left = self.left if isinstance(self.left, Statement) else self.dialect.placeholder
        right = self.right if isinstance(self.right, Statement) else self.dialect.placeholder

        return f"({left!s} {self.operator} {right!s})"

    @property
    def args(self) -> list[Any]:
        out: list[Any] = []

        if isinstance(self.left, Statement):
            out.extend(self.left.args)
        else:
            out.append(self.left)

        if isinstance(self.right, Statement):
            out.extend(self.right.args)
        else:
            out.append(self.right)

        return out


class RawExpression(Expression, Raw):
    """Expression as result of another expression."""


class Column(Expression):
    """
    Column (optionally with table and database) as statement.

    >>> from sqlfactory import Column
    >>> Column("column")
    >>> "`column`"

    >>> from sqlfactory import Column
    >>> Column("table.column")
    >>> "`table`.`column`"

    >>> from sqlfactory import Column
    >>> Column("database.table.column")
    >>> "`database`.`table`.`column`"

    Star-columns are also supported, although they are not recommended to be used. Always try to specify exact columns to select.

    >>> from sqlfactory import Column
    >>> Column("*")
    >>> "*"

    >>> from sqlfactory import Column
    >>> Column("table.*")
    >>> "`table`.*"

    >>> from sqlfactory import Column
    >>> Column("database.table.*")
    >>> "`database`.`table`.*"

    To use column with `Select` statement with alias, you can use `Aliased` or `SelectColumn` classes instead.

    The class also provides shorthand for creating conditional SQL statements and arithmetic SQL statements. You can use Column
    instance to directly create condition using simple operators (`==`, `!=`, `<`, `<=`, `>`, `>=`) or arithmetic operations
    (`+`, `-`, `*`, `/`, `%`).

    >>> from sqlfactory import Column
    >>> Column("table.column") == 5
    >>> # Produces Eq(Column("table.column"), 5)

    >>> from sqlfactory import Column
    >>> Column("table.column") + 5
    >>> # Produces Expression("`table`.`column` + %s", 5)
    """

    def __init__(self, column: str) -> None:
        super().__init__()

        self._column = column.split(".")
        if len(self._column) > 3:
            raise ValueError("Invalid column name (contains more than <database>.<table>.<column>).")

    def __str__(self) -> str:
        return ".".join(
            f"{self.dialect.quote}{x}{self.dialect.quote}" if not x.startswith(self.dialect.quote) and x != "*" else x
            for x in self._column
        )

    @property
    def column(self) -> str:
        """Returns column part of the column name."""
        return self._column[-1]

    @property
    def table(self) -> str | None:
        """Returns table part of the column name, if specified. If column specification does not contain table name,
        returns None."""
        try:
            return self._column[-2]
        except IndexError:
            return None

    @property
    def database(self) -> str | None:
        """Returns database part of the column name, if specified. If column specification does not contain database
        name, returns None."""
        try:
            return self._column[-3]
        except IndexError:
            return None

    def __hash__(self) -> int:
        return hash(str(self))

    @property
    def args(self) -> list[Any]:
        """Column does not have any arguments, so this always returns empty list."""
        return []


class Table(Statement):
    """
    Table (optionally with database) as statement

    >>> from sqlfactory import Table
    >>> Table("table")
    >>> "`table`"

    >>> from sqlfactory import Table
    >>> Table("database.table")
    >>> "`database`.`table`"

    To produce table alias, use `Aliased` class:

    >>> from sqlfactory import Table, Aliased
    >>> Aliased(Table("database.table"), alias="t1")
    >>> "`database`.`table` AS `t1`"

    By accessing Table's undefined attributes, instance of Column is returned. This allows to easily access columns of
    that table, and reference them in SQL statements.

    >>> from sqlfactory import Table, Select
    >>> t = Table("table")
    >>> Select(t.id, t.name, table=t)
    >>> "SELECT `table`.`id`, `table`.`name` FROM `table`"

    This allows creating python-like expressions that are converted to SQL automatically:

    >>> from sqlfactory import Table, Select
    >>> from sqlfactory.func.agg import Count
    >>> t = Table("database.table")
    >>> Select(Count(t.id), table=t, where=t.id > 5)
    >>> "SELECT COUNT(`database`.`table`.`id`) FROM `database`.`table` WHERE `database`.`table`.`id` > %s"
    """

    def __init__(self, table: str) -> None:
        """
        :param table: Table name (optionally with database in form `<database>.<table>`). If database is not specified,
            it is assumed that table is in default database and SQL constructed will not contain any database name.
        """
        super().__init__()
        self._table = table.split(".")

        if len(self._table) > 2:
            raise ValueError("Invalid table name (contains more than <database>.<table>).")

    def __str__(self) -> str:
        return ".".join(
            f"{self.dialect.quote}{x}{self.dialect.quote}" if not x.startswith(self.dialect.quote) else x for x in self._table
        )

    @property
    def table(self) -> str | None:
        """Returns table part of the table name"""
        return self._table[-1]

    @property
    def database(self) -> str | None:
        """Returns database part of the table name, if specified. If table specification does not contain database,
        returns None."""
        try:
            return self._table[-2]
        except IndexError:
            return None

    def __getattr__(self, name: str) -> Column:
        """Returns column of that table."""
        return Column(f"{'.'.join(self._table)}.{name}")

    @property
    def __isabstractmethod__(self) -> bool:
        """
        This is needed to be able to use Table instance in abstract classes as ClassVar.

        class Something(ABC):
            table: ClassVar[Table] = Table("table")

        - ABC checks for __isabstractmethod__ property for all attributes in __dict__, and because Table did not
        contains it, __getattr__ is called, which returns Column, which evaluates to True, which in turn makes
        ABC think that Table is abstract method. So by providing explicit implementation here, we avoid this error.
        """
        return False

    @property
    def args(self) -> list[Any]:
        """Table does not have any arguments, this always returns empty list."""
        return []


# Alias for column that can be passed as instance of column or as string.
ColumnArg = Column | str



"""UPDATE statement builder."""

from __future__ import annotations

from typing import Any, Collection, Mapping, Optional, Self, TypeAlias

from sqlfactory.condition.base import ConditionBase
from sqlfactory.dialect import SQLDialect
from sqlfactory.entities import Column, ColumnArg, Table
from sqlfactory.execute import ConditionalExecutableStatement
from sqlfactory.mixins.join import Join, WithJoin
from sqlfactory.mixins.limit import Limit, WithLimit
from sqlfactory.mixins.order import OrderArg, WithOrder
from sqlfactory.mixins.where import WithWhere
from sqlfactory.statement import Statement


class UpdateColumn(Statement):
    """
    Represents one field that should be updated.
    """

    def __init__(self, column: ColumnArg, value: Statement | Any) -> None:
        super().__init__()

        self._column = column if isinstance(column, Column) else Column(column)
        self._value = value

    def __str__(self) -> str:
        return f"{self._column!s} = {str(self._value) if isinstance(self._value, Statement) else self.dialect.placeholder}"

    def __hash__(self) -> int:
        return hash(self._column)

    def __eq__(self, other: UpdateColumn | Any) -> bool:
        if not isinstance(other, UpdateColumn):
            return False

        return str(self._column) == str(other._column)

    @property
    def args(self) -> list[Any]:
        """
        Return arguments for the update statement.
        """
        if isinstance(self._value, Statement):
            return self._value.args

        return [self._value]


class Update(ConditionalExecutableStatement, WithWhere, WithLimit, WithOrder, WithJoin):
    # pylint: disable=too-many-ancestors
    """
    Builds `UPDATE` statement SQL query.

    This is conditional SQL statement, so you can check whether it should be executed (would update any columns)
    by calling bool() on it. Also, if you are using execute() method of the statement, the execution won't be
    performed if bool() returns False.

    Examples:

    >>> from sqlfactory import Update, Eq
    >>> upd = Update("table").set("column1", 1).where(Eq("column2", 2))
    >>> str(upd)
    'UPDATE `table` SET `column1` = %s WHERE `column2` = %s'
    >>> upd.args
    [1, 2]

    >>> from sqlfactory import Update, Table
    >>> t = Table("table")
    >>> upd = Update(t).set(t.column1, 1).where(t.column2 == 2)
    >>> str(upd)
    'UPDATE `table` SET `table`.`column1` = %s WHERE `table`.`column2` = %s'
    >>> upd.args
    [1, 2]

    >>> from sqlfactory import Update
    >>> upd = Update("table", set={"column1": 1}, where=Eq("column2", 2))
    >>> str(upd)
    'UPDATE `table` SET `column1` = %s WHERE `column2` = %s'
    >>> upd.args
    [1, 2]

    >>> from sqlfactory import Update, UpdateColumn
    >>> upd = Update("table", UpdateColumn("column1", 1), where=Eq("column2", 2))
    >>> str(upd)
    'UPDATE `table` SET `column1` = %s WHERE `column2` = %s'
    >>> upd.args
    [1, 2]

    >>> from sqlfactory import Update
    >>> upd = Update(
    ...     ["table1", "table2"],
    ...     set={"table1.value": "foo", "table2.value": "bar"},
    ...     where=Eq("table1.id", Column("table2.id"))
    ... )
    >>> str(upd)
    'UPDATE `table1`, `table2` SET `table1`.`value` = %s, `table2`.`value` = %s WHERE `table1`.`id` = `table2`.`id`'
    >>> upd.args
    ['foo', 'bar']

    >>> from sqlfactory import Update, Join, Eq
    >>> upd = Update(
    ...     "table1",
    ...     join=[Join("table2", Eq("table1.id", "table2.id"))],
    ...     set={"table1.value": "foo", "table2.value": "bar"}
    ... )
    >>> str(upd)
    'UPDATE `table1` JOIN `table2` ON `table1`.`id` = `table2`.`id` SET `table1`.`value` = %s, `table2`.`value` = %s'
    >>> upd.args
    ['foo', 'bar']

    """

    def __init__(
        self,
        table: Table | str | Collection[Table | str],
        *fields: UpdateColumn,
        join: Collection[Join] | None = None,
        set: Optional[Mapping[str | Column, Any]] = None,  # noqa: A002
        where: Optional[ConditionBase] = None,
        order: OrderArg | None = None,
        limit: Optional[Limit] = None,
        dialect: SQLDialect | None = None,
    ) -> None:
        # pylint: disable=redefined-builtin
        """
        :param table: Table to update. Or multiple tables, if collection is used.
        :param fields: List of UpdateColumn instances containing columns to be updated. This is not very pleasant way
            to create the statement, use set() method instead.
        :param join: Optional join to update multiple tables.
        :param set: Mapping of column names to values to be set. Alternative to specifying columns as UpdateColumn variable
            arguments. You can also use set() method on the Update instance to add another columns to be updated.
        :param where: WHERE condition
        :param order: Optional ordering of the updated rows. Might be needed to avoid violating keys when updating multiple rows.
        :param limit: Limit number of updated rows
        """
        super().__init__(join=join, where=where, order=order, limit=limit, dialect=dialect)

        if isinstance(table, (Table, str)):
            table = [table]

        self.table = [t if isinstance(t, Table) else Table(t) for t in table]
        """Tables that should be updated."""

        self.fields: list[UpdateColumn] = list(fields)
        """Fields that should be updated."""

        if set:
            for column, value in set.items():
                self.set(column, value)

    def __str__(self) -> str:
        """
        Return the UPDATE statement with %s placeholders for arguments.
        """
        with self.dialect:
            if not self.fields:
                raise AttributeError("At least one column must be updated.")

            query = [f"UPDATE {', '.join(str(t) for t in self.table)}"]

            if self._join:
                query.extend(map(str, self._join))

            query.append(f"SET {', '.join(map(str, self.fields))}")

            if self._where:
                query.append("WHERE")
                query.append(str(self._where))

            if self._order:
                query.append(str(self._order))

            if self._limit:
                query.append(str(self._limit))

            return " ".join(query)

    @property
    def args(self) -> list[Any]:
        """
        Return all arguments used in the UPDATE statement.
        """
        out = []

        if self._join:
            for join in self._join:
                out.extend(join.args)

        for field in self.fields:
            out.extend(field.args)

        if self._where is not None:
            out.extend(self._where.args)

        if self._order is not None:
            out.extend(self._order.args)

        if self._limit is not None:
            out.extend(self._limit.args)

        return out

    def __bool__(self) -> bool:
        """
        Return True if there are any fields to be updated and the statement should be executed.
        """
        return bool(self.fields)

    def append(self, field: UpdateColumn) -> Self:
        """
        Append new UpdateField to this UPDATE statement. Can be used when set() method is not sufficient.
        """
        if field in self.fields:
            raise AttributeError(f"Field '{field}' is already in the list of fields to be updated.")

        self.fields.append(field)
        return self

    def set(self, field: ColumnArg, value: Statement | Any) -> Self:
        """
        Syntactical sugar for creating simple SET UpdateFields.
        :param field: Field name (without quotes).
        :param value: Value to set the field to (will be escaped).
        """
        return self.append(UpdateColumn(field, value))

    def SET(self, field: str, value: Any) -> Self:
        # pylint: disable=invalid-name
        """Alias for `Update.set()` for better SQL compatibility (SQL is often written in all caps)."""
        return self.set(field, value)


UPDATE: TypeAlias = Update  # pylint: disable=invalid-name
"""
Alias for `Update` statement to provide better SQL compatibility, as SQL is often written in all caps.
"""
